/**
 * \file   builddeb.c
 * \brief  Program entry point.
 * \author Copyright (c) 2010 pasgui
 *
 * Hash: $Format:%h$
 * Date: $Format:%ci$
 *
 */

#include <stdlib.h>
#include <string.h>
#include "builddeb.h"
#include "lfs.h"
#include "lbz.h"
#include "lsqlite3.h"
#include "lxmlreader.h"

#define VERSION        "1.2.7"
#define COPYRIGHT      "Copyright (c) 2010-2014 pasgui"
#define ERROR_MESSAGE  "%s\n"
#define WARNING_MESSAGE  "%s\n"

static int process_arguments(lua_State* L, int argc, const char** argv);
static int process_option(lua_State* L, const char* arg);
static int load_builtin_scripts(lua_State* L);

/* A search path for script files */
static const char* scripts_path = NULL;

/* precompiled bytecode buffer; in bytecode.c */
extern const char* builtin_scripts[];

/**
 * Program entry point.
 */
int main(int argc, const char** argv)
{
  lua_State* L;
  int z = OKAY;

  /* prepare Lua for use */
  L = luaL_newstate();
  luaL_openlibs(L);
  luaopen_lbz2(L);
  luaopen_lsqlite3(L);
  luaopen_lfs(L);
  luaopen_xmlreader(L); 

  /* push the application metadata */
  lua_pushstring(L, LUA_COPYRIGHT);
  lua_setglobal(L, "_COPYRIGHT");

  lua_pushstring(L, VERSION);
  lua_setglobal(L, "_BUILDDEB_VERSION");

  lua_pushstring(L, COPYRIGHT);
  lua_setglobal(L, "_BUILDDEB_COPYRIGHT");

  /* Parse the command line arguments */
  if (z == OKAY)  z = process_arguments(L, argc, argv);

  /* Run the built-in Premake scripts */
  if (z == OKAY)  z = load_builtin_scripts(L);

  /* Clean up and turn off the lights */
  lua_close(L);
  return z;
}


/**
 * Process the command line arguments, splitting them into options, the
 * target action, and any arguments to that action. The results are pushed
 * into the session for later use. I could have done this in the scripts,
 * but I need the value of the /scripts option to find them.
 * \returns OKAY if successful.
 */
int process_arguments(lua_State* L, int argc, const char** argv)
{
  int i;
  
  /* Create empty lists for Options and Args */
  lua_newtable(L);
  lua_newtable(L);

  for (i = 1; i < argc; ++i)
  {
    /* Options start with '/' or '--'. The first argument that isn't an option
     * is the action. Anything after that is an argument to the action */
    if (argv[i][0] == '/')
    {
      process_option(L, argv[i] + 1);
    }
    else if (argv[i][0] == '-' && argv[i][1] == '-')
    {
      process_option(L, argv[i] + 2);
    }
    else
    {
      /* not an option, is the action */
      lua_pushstring(L, argv[i++]);
      lua_setglobal(L, "_ACTION");

      /* everything else is an argument */
      while (i < argc)
      {
        lua_pushstring(L, argv[i++]);
        lua_rawseti(L, -2, lua_rawlen(L, -2) + 1);
      }
    }
  }

  /* push the Options and Args lists */
  lua_setglobal(L, "_ARGS");
  lua_setglobal(L, "_OPTIONS");
  return OKAY;
}



/**
 * Parse an individual command-line option.
 * \returns OKAY if successful.
 */
int process_option(lua_State* L, const char* arg)
{
  char key[512];
  const char* value;

  /* If a value is specified, split the option into a key/value pair */
  char* ptr = strchr(arg, '=');
  if (ptr)
  {
    int len = ptr - arg;
    if (len > 511) len = 511;
    strncpy(key, arg, len);
    key[len] = '\0';
    value = ptr + 1;
  }
  else
  {
    strcpy(key, arg);
    value = "";
  }

  /* Store it in the Options table, which is already on the stack */
  lua_pushstring(L, value);
  lua_setfield(L, -3, key);

  /* The /scripts option gets picked up here to find the built-in scripts */
  if (strcmp(key, "scripts") == 0 && strlen(value) > 0)
  {
    scripts_path = value;
  }

  return OKAY;
}



#if defined(_DEBUG)
/**
 * When running in debug mode, the scripts are loaded from the disk. The path to
 * the scripts must be provided via either the /scripts command line option or
 * the BUILDDEB_PATH environment variable.
 */
int load_builtin_scripts(lua_State* L)
{
  const char* filename;
  int i;
  const char * path[3] = {scripts_path, getenv("BUILDDEB_PATH"), "src"};

  /* search the builddeb.lua file */  
  for(i=0; i<3; i++)
  {
    if(path[i] != NULL)
    {
      lua_getglobal(L, "lfs");
      lua_getfield(L, -1, "attributes");
      lua_remove(L, -2);
      lua_pushfstring(L, "%s/_builddeb.lua", path[i]);
      if (lua_pcall(L, 1, 1, 0) != OKAY)
      {
        printf(ERROR_MESSAGE, lua_tostring(L, -1));
        return !OKAY;
      }
      if (lua_istable(L, -1))
      {
        break;
      }
    }
  }
  if(i == 3)
  {
    printf(WARNING_MESSAGE, 
      "Unable to find builddeb.lua; use /scripts option when in debug mode!\n"
      "Please refer to the documentation (or build in release mode instead)."
    );
    return !OKAY;
  }

  /* run the bootstrapping script */
  filename = lua_pushfstring(L, "%s/_builddeb.lua", path[i]);
  if (luaL_dofile(L, filename))
  {
    printf(ERROR_MESSAGE, lua_tostring(L, -1));
    return !OKAY;
  }

  /* hand off control to the scripts */
  lua_getglobal(L, "_builddeb_main");
  lua_pushstring(L, path[i]);
  if (lua_pcall(L, 1, 1, 0) != OKAY)
  {
    printf(ERROR_MESSAGE, lua_tostring(L, -1));
    return !OKAY;
  }
  else
  {
    return (int)lua_tonumber(L, -1);
  }
}
#endif


#if defined(NDEBUG)
/**
 * When running in release mode, the scripts are loaded from a static data
 * buffer, where they were stored as string in a table. To regenerate
 * the table, run `premake4 embed` and then rebuild.
 */
int load_builtin_scripts(lua_State* L)
{
  int i;
  for (i = 0; builtin_scripts[i]; ++i)
  {
    if (luaL_dostring(L, builtin_scripts[i]) != OKAY)
    {
      printf(ERROR_MESSAGE, lua_tostring(L, -1));
      return !OKAY;
    }
  }

  /* hand off control to the scripts */
  lua_getglobal(L, "_builddeb_main");
  if (lua_pcall(L, 0, 1, 0) != OKAY)
  {
    printf(ERROR_MESSAGE, lua_tostring(L,-1));
    return !OKAY;
  }

  return OKAY;
}
#endif
